'''
Created on 24 juin 2023

@author: tleduc

Copyright 2020-2023 Thomas Leduc

This file is part of t4gpd.

t4gpd is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

t4gpd is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with t4gpd.  If not, see <https://www.gnu.org/licenses/>.
'''
from geopandas import clip, GeoDataFrame
from datetime import datetime, timedelta, timezone
import pyvista
from numpy import asarray
from pyvista import Cylinder, Disc, global_theme, Light, Plane, Plotter, Sphere, themes
from t4gpd.commons.ColorTemperature import ColorTemperature
from t4gpd.commons.GeomLib import GeomLib
from t4gpd.commons.GeoProcess import GeoProcess
from t4gpd.commons.LatLonLib import LatLonLib
from t4gpd.commons.sun.SunLib import SunLib
from t4gpd.demos.GeoDataFrameDemos import GeoDataFrameDemos
from t4gpd.morph.geoProcesses.FootprintExtruder import FootprintExtruder
from t4gpd.morph.geoProcesses.STGeoProcess import STGeoProcess
from t4gpd.pyvista.ToPolyData import ToPolyData
from t4gpd.pyvista.ToUnstructuredGrid import ToUnstructuredGrid
from zoneinfo import ZoneInfo


class Plotter3D(GeoProcess):
    '''
    classdocs
    '''

    def __init__(self, window_size=(1100, 1100), theme=themes.ParaViewTheme(), ofile=None):
        '''
        Constructor
        '''
        self.ofile = ofile

        self.plotter = Plotter(off_screen=(ofile is not None), lighting="none",
                               window_size=window_size)
        self.plotter.show_axes()
        # pyvista.set_plot_theme(themes.DarkTheme())
        # pyvista.set_plot_theme(themes.DocumentTheme())
        # pyvista.set_plot_theme(themes.ParaViewTheme())
        pyvista.set_plot_theme(theme)

    def add_mesh(self, scene, color=None, opacity=1.0, show_edges=False,
                 show_scalar_bar=False):
        self.plotter.add_mesh(scene, color=color, opacity=opacity, show_edges=show_edges,
                              show_scalar_bar=show_scalar_bar, smooth_shading=True)

    def add_buildings(self, buildings, color="grey", opacity=1.0, show_edges=False,
                      show_scalar_bar=False, elevationFieldname='HAUTEUR',
                      forceZCoordToZero=True, withoutBuildingFloors=False):
        _buildings = buildings.explode(ignore_index=True)

        op = FootprintExtruder(_buildings, elevationFieldname,
                               forceZCoordToZero, withoutBuildingFloors)
        buildingsIn3d = STGeoProcess(op, _buildings).run()

        # scene = ToUnstructuredGrid([buildingsIn3d], fieldname=None).run()
        scene = ToPolyData([buildingsIn3d], fieldname=None).run()
        self.add_mesh(scene, color=color, show_edges=show_edges,
                      show_scalar_bar=show_scalar_bar)

    def add_ground_surface(self, length=None, npts=64, color=None, opacity=1.0,
                           show_edges=False, show_scalar_bar=False):
        xmin, xmax, ymin, ymax, zmin, zmax = self.plotter.bounds
        xc, yc, zc = self.plotter.center

        if length is None:
            length = self.plotter.length
            ground = Plane(center=[xc, yc, zmin],
                           i_size=length, j_size=length,
                           i_resolution=10, j_resolution=10)
        else:
            ground = Disc(center=[xc, yc, zmin], inner=0.0,
                          outer=length, normal=(0.0, 0.0, 1.0),
                          r_res=1, c_res=npts)

        self.add_mesh(ground, color=color, show_edges=show_edges,
                      show_scalar_bar=show_scalar_bar)

    def add_general_lighting(self, color="grey", intensity=0.95):
        xmin, xmax, ymin, ymax, zmin, zmax = self.plotter.bounds
        xc, yc, zc = self.plotter.center
        light = Light(
            position=(xc, yc, zmax+100),
            focal_point=(xc, yc, zc),
            color=color,  # "blue"
            intensity=intensity
        )
        self.plotter.add_light(light)

    def add_streetlights(self, streetlights, generalLighting=True,
                         color=[1.0, 0.83921, 0.6666, 1.0],
                         aperture=75.5, show_actor=False):
        # https://docs.pyvista.org/version/stable/examples/04-lights/actors.html
        # https://docs.pyvista.org/version/stable/api/core/_autosummary/pyvista.Light.html#pyvista.Light
        # https://docs.pyvista.org/version/stable/examples/04-lights/shadows.html
        for _, row in streetlights.iterrows():
            x0, y0, z0 = row.geometry.coords[0]

            if isinstance(color, str):
                color = row[color]
            if isinstance(aperture, str):
                aperture = row[aperture]

            # A positional light with a cone angle that is less than 90 degrees
            # is known as a spotlight
            light = Light(
                position=(x0, y0, z0),
                focal_point=(x0, y0, z0-1),
                color=color,
                intensity=1,  # brightness of the light (between 0 and 1)
                positional=True,
                cone_angle=aperture,
                exponent=10,
                show_actor=show_actor
            )
            self.plotter.add_light(light)

        if generalLighting:
            self.add_general_lighting(color="dimgrey", intensity=0.35)

    def add_sun_shadows(self, dts, gdf=LatLonLib.NANTES, model="pysolar", colors=None):
        # color = [1.0, 1.0, 0.9843, 1.0],  # Color temp. 5400 K
        # color = [1.0, 0.83921, 0.6666, 1.0],  # Color temp. 2850 K
        self.plotter.enable_shadows()
        sunModel = SunLib(gdf, model)

        if colors is None:
            colors = [ColorTemperature.convert_K_to_RGB(
                2850 + 1000*i) for i in range(len(dts))]

        for i, dt in enumerate(dts):
            solarAlti, solarAzim = sunModel.getSolarAnglesInDegrees(dt)
            light = Light(
                color=colors[i],
                positional=False,
            )
            light.set_direction_angle(solarAlti, solarAzim)
            self.plotter.add_light(light)
            print(f"{dt}: alti={solarAlti:.1f}, azim={solarAzim:.1f}")

    def add_sun_positions(self, dts, radius=10, length=None,
                          gdf=LatLonLib.NANTES, model="pysolar",
                          centre=None, color="yellow", opacity=1.0,
                          show_edges=False, show_scalar_bar=False):
        xc, yc, zc = self.plotter.center if centre is None else centre
        length = self.plotter.length/2 if length is None else length
        sunModel = SunLib(gdf, model)

        for i, dt in enumerate(dts):
            x, y, z = sunModel.getRadiationDirection(dt)
            xyz = xc + length*x, yc + length*y, zc + length*z
            sun = Sphere(radius=radius, center=xyz)
            self.add_mesh(sun, color=color, show_edges=show_edges,
                          show_scalar_bar=show_scalar_bar)
        return xc, yc, zc

    def add_tree_cylinders(self):
        pass

    def add_tree_spheres(self, trees, trunk_height="trunk_height",
                         crown_radius="crown_radius", color="green",
                         opacity=1.0, show_edges=False,
                         show_scalar_bar=False):
        for _, row in trees.iterrows():
            h, r = row[trunk_height], row[crown_radius],
            xyz = list(row.geometry.coords[0])[0:2] + [h]

            houppier = Sphere(radius=r, center=xyz)
            tronc = Cylinder(center=xyz[0:2] + [xyz[2]/2],
                             direction=(0.0, 0.0, 1.0),
                             radius=0.2, height=h)
            self.add_mesh(houppier, color=color, opacity=1.0,
                          show_edges=False,
                          show_scalar_bar=show_scalar_bar)
            self.add_mesh(tronc, color=color, opacity=1.0,
                          show_edges=False,
                          show_scalar_bar=show_scalar_bar)

    def my_cpos_callback(self):
        # tuple: camera location, focus point, viewup vector
        # camera_position = [(x,y,z), (fx,fy,fz,), (nx,ny,nz)]
        cpos = self.plotter.camera_position
        (x, y, z), (fx, fy, fz,), (nx, ny, nz) = \
            cpos.position, cpos.focal_point, cpos.viewup
        msg = f"""[({x:.2f}, {y:.2f}, {z:.2f}),
 ({fx:.2f}, {fy:.2f}, {fz:.2f}),
 ({nx:.2f}, {ny:.2f}, {nz:.2f})]
"""
        # self.plotter.add_text(str(self.plotter.camera_position), name="cpos")
        self.plotter.add_text(msg, name="cpos")
        print(msg)
        return

    def set_camera_position(self, cpos=None):
        # tuple: camera location, focus point, viewup vector
        # self.plotter.camera_position = [(x,y,z), (fx,fy,fz,), (nx,ny,nz)]
        if not cpos is None:
            self.plotter.camera_position = cpos

    def run(self):
        self.plotter.add_key_event("p", self.my_cpos_callback)
        self.plotter.show(screenshot=self.ofile)


"""
tzinfo = ZoneInfo("Europe/Paris")

buildings = GeoDataFrameDemos.ensaNantesBuildings()
trees = GeoDataFrameDemos.ensaNantesTrees()
trees["trunk_height"] = 5.0
trees["crown_radius"] = 3.0

streetlights = GeoDataFrameDemos.ensaNantesTrees()
streetlights.geometry = streetlights.geometry.apply(
    lambda g: GeomLib.forceZCoordinateToZ0(g, 8.0))
streetlights = streetlights.head(6)

# ofile = "/tmp/test.png"
ofile = None
pl = Plotter3D(ofile)
pl.add_buildings(buildings, show_edges=False)
pl.add_tree_spheres(trees)
# pl.add_streetlights(streetlights, generalLighting=True)
pl.add_ground_surface()

dt0 = datetime(2021, 3, 21, 10, tzinfo=tzinfo)
pl.add_sun_shadows([dt0, dt0 + timedelta(hours=6)])

# dt0 = datetime(2021, 3, 21, 8, tzinfo=tzinfo)
# dts = [dt0 + timedelta(hours=i) for i in range(0, 12, 2)]
# pl.add_sun_positions(dts)
pl.run()
"""

'''
tzinfo = timezone.utc

buffDist = 140.0
roi = GeoDataFrameDemos.coursCambronneInNantes()
roi = roi.geometry.squeeze().centroid.buffer(buffDist)
buildings = GeoDataFrameDemos.districtGraslinInNantesBuildings()
buildings = clip(buildings, roi, keep_geom_type=True)

ground = GeoDataFrame(
    [{"geometry": GeomLib.forceZCoordinateToZ0(roi, z0=0)}], crs="epsg:2154")

trees = GeoDataFrameDemos.districtGraslinInNantesTrees()
trees = clip(trees, roi, keep_geom_type=True)
trees["trunk_height"] = 5.0
trees["crown_radius"] = 3.0

# ofile = "/tmp/test.png"
ofile = None
# pl = Plotter3D(ofile, theme=themes.ParaViewTheme())
pl = Plotter3D(ofile, theme=themes.DocumentTheme())
# pl = Plotter3D(ofile, theme=themes.DarkTheme())

pl.add_buildings(buildings, show_edges=False)
pl.add_tree_spheres(trees)
pl.add_ground_surface(buffDist)

dt0 = datetime(2021, 6, 21, 4, tzinfo=tzinfo)
dts = [dt0 + timedelta(hours=i) for i in range(0, 17, 1)]
pl.add_sun_positions(dts, radius=10, length=160)

# dt0 = datetime(2021, 3, 21, 9, tzinfo=tzinfo)
# dts = [dt0 + timedelta(hours=i) for i in range(1, 6, 1)]
# pl.add_sun_shadows(dts)

cpos = [(354975.59, 6689462.60, 435.57),
        (354805.12, 6689029.84, 16.62),
        (-0.30, -0.60, 0.74)]
pl.set_camera_position(cpos)
pl.my_cpos_callback()
pl.add_general_lighting(color="lightgrey", intensity=0.95)
pl.run()
'''



'''
Created on 6 mai 2023

@author: tleduc

Copyright 2020-2023 Thomas Leduc

This file is part of t4gpd.

t4gpd is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

t4gpd is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with t4gpd.  If not, see <https://www.gnu.org/licenses/>.
'''
import warnings

from numpy import isnan
from t4gpd.commons.GeoProcess import GeoProcess
from t4gpd.commons.WarnUtils import WarnUtils
from t4gpd.picoclim.InertialMeasureReader import InertialMeasureReader


class InertialMeasureReWriter(GeoProcess):
    '''
    classdocs
    '''
    OK = 0
    CANNOT_ADD_FIRST_TN = 1 
    CANNOT_ADD_LAST_TN = 2 
    
    def __init__(self, inputFile, outputFile, rewriteTagNames=True, addFirstTagName=True, addLastTagName=True):
        '''
        Constructor
        '''
        warnings.formatwarning = WarnUtils.format_Warning
        self.inputFile = inputFile
        self.outputFile = outputFile
        self.rewriteTagNames = rewriteTagNames
        self.addFirstTagName = addFirstTagName
        self.addLastTagName = addLastTagName

    def __read_csv(self):
        version = InertialMeasureReader.getVersion(self.inputFile)
        # with open(self.inputFile, "r") as ifile:
        #     head = [next(ifile) for _ in (0, 1)]
        head = """     time_stamp step_count           X       Y       Z    Distance          degree          latitude       longitude        GpsAccuracy  indoor_outdoor_flag       TagName

"""
        df = InertialMeasureReader(self.inputFile).run()
        return version, head, df

    @staticmethod
    def tagNameDiagnosis(inputFile):
        df = InertialMeasureReader(inputFile).run()
        exit_status = InertialMeasureReWriter.OK
        tagNames = df[~df.TagName.isna()].TagName.to_dict()
        trackId = {(v // 10) if (v < 100) else (v // 100): None for v in tagNames.values()}
        if (1 != len(trackId)):
            raise Exception(f"The TagNames' fall under more than one track in {inputFile}!")
        trackId = int(list(trackId.keys())[0])
        if ((0 in tagNames) or (1 in tagNames)):
            # warnings.warn("I can't insert a starting TagName!")
            exit_status += InertialMeasureReWriter.CANNOT_ADD_FIRST_TN
        n = len(df) - 1
        if (((n - 1) in tagNames) or (n in tagNames)):
            # warnings.warn("I can't insert an end TagName!")
            exit_status += InertialMeasureReWriter.CANNOT_ADD_LAST_TN
        return trackId, exit_status
        
    @staticmethod
    def rectifyId(i):
        if (i < 100):
            tens, units = i // 10, i % 10
            return tens * 100 + units
        return i

    def __rewriteTagNames(self, df):
        exit_status = self.OK
        oldTagNames = df[~df.TagName.isna()].TagName.to_dict()
        newTagNames = {k: self.rectifyId(v) for k, v in oldTagNames.items()}

        trackId = {v // 100: None for v in newTagNames.values()}
        if (1 != len(trackId)):
            raise Exception(f"The TagNames' fall under more than one track in {self.inputFile}!")
        trackId = int(list(trackId.keys())[0])

        # HANDLE 31 TRACKS #4 IN 06-Carriere-misery-Nantes (id: 40 -> 42)
        if (0 == min(newTagNames.values()) % 100):
            newTagNames = {k: v+1 for k, v in newTagNames.items()}

        if self.addFirstTagName:
            if ((0 in newTagNames) or (1 in newTagNames)):
                # warnings.warn("I can't insert a starting TagName!")
                exit_status += self.CANNOT_ADD_FIRST_TN
            else:
                newTagNames[0] = trackId * 100
        if self.addLastTagName:
            n = len(df) - 1
            if (((n - 1) in newTagNames) or (n in newTagNames)):
                # warnings.warn("I can't insert an end TagName!")
                exit_status += self.CANNOT_ADD_LAST_TN
            else:
                newTagNames[n] = 1 + max(newTagNames.values())
        newTagNames = {k:v for k, v in sorted(newTagNames.items(), key=lambda kv: kv[0])}

        for k, v in newTagNames.items():
            df.at[k, "TagName"] = v

        # print(oldTagNames)
        # print(newTagNames)
        return exit_status, df

    def __to_csv(self, version, head, df):
        # df.to_csv(self.outputFile)
        # FMT = "{:>15s}{:>10d}{:>13.2f}{:>8.2f}{:>8.2f}{:>12.2f}{:>16.2f}{:>18.6f}{:>16.6f}{:>19.2f}{:>21s}{:>14.0f}\n"
        FMT = "{:>15s}{:>10d}{:>13s}{:>8s}{:>8s}{:>12s}{:>16s}{:>18s}{:>16s}{:>19s}{:>21s}{:>14s}\n"

        foo1 = lambda dt: dt[:-3]
        foo2 = lambda x: f"{x:.2f}".replace(".", ",")
        foo3 = foo2
        foo4 = lambda x: f"{x:.6f}".replace(".", ",")
        foo5 = lambda x: x.strip()
        foo6 = lambda x: "" if isnan(x) else f"{x:.0f}"

        if (2 == version):
            # FMT = "{:>11s}{:>14d}{:>13s}{:>8s}{:>8s}{:>12s}{:>16s}{:>16s}{:>16s}{:>21s}{:>21s}{:>12s}\n"
            foo1 = lambda dt: dt[:-7]
            foo4 = foo2
        elif (4 == version):
            # FMT = "{:<16s}{:>4d}{:>13s}{:>8s}{:>8s}{:>13s}{:>16s}{:>11s}{:>15s}{:>14s}{:>11s}{:>26s}\n"
            # FMT = "{:>11s}{:>14d}{:>13s}{:>8s}{:>8s}{:>12s}{:>16s}{:>16s}{:>16s}{:>21s}{:>21s}{:>12s}\n"
            foo3 = lambda x: x.replace(".", ",")

        with open(self.outputFile, "w") as ofile:
            ofile.writelines(head)

            for _, row in df.iterrows():
                line = FMT.format(
                    foo1(row.timestamp.strftime("%H:%M:%S.%f")),
                    row.step_count, foo2(row.X), foo2(row.Y), foo2(row.Z),
                    foo3(row.Distance), foo2(row.degree), foo4(row.latitude), foo4(row.longitude),
                    foo2(row.GpsAccuracy), foo5(row.indoor_outdoor_flag), foo6(row.TagName)
                )
                ofile.write(line)

    def run(self):
        exit_status = self.OK
        version, head, df = self.__read_csv()

        if (self.rewriteTagNames) or (self.addFirstTagName) or (self.addLastTagName):
            exit_status, df = self.__rewriteTagNames(df)

        self.__to_csv(version, head, df)
        print(f"{self.outputFile} has been written (v{version})!")
        return exit_status

"""
inputFile = "/home/tleduc/prj/coolscapes-tl/dev/data/05-Quai-de-plantes-Nantes/20200526/localisation/LOG_FILE_20200526_120307.txt"
InertialMeasureReWriter(inputFile, "/tmp/abc.csv").run()

inputFile = "/home/tleduc/prj/coolscapes-tl/dev/data/06-Carriere-misery-Nantes/20200722/localisation/LOG_FILE_20200722_141610.txt"
InertialMeasureReWriter(inputFile, "/tmp/abc.csv").run()

inputFile = "/home/tleduc/prj/coolscapes-tl/dev/data/08-La-defense-Paris/20210721/localisation/LOG_FILE_20210721_182401.txt"
InertialMeasureReWriter(inputFile, "/tmp/abc.csv").run()

inputFile = "/home/tleduc/prj/nm-ilots-frais/terrain/220711/LOG_FILE_20220711_172000.txt"
InertialMeasureReWriter(inputFile, "/tmp/abc.csv").run()

inputFile = "/home/tleduc/prj/coolscapes-tl/dev/data/06-Carriere-misery-Nantes/20200716/localisation/LOG_FILE_20200716_115024.txt"
InertialMeasureReWriter(inputFile, "/tmp/abc.csv", rewriteTagNames=True, addFirstTagName=False, addLastTagName=False).run()
"""



'''
Created on 23 nov. 2020

@author: tleduc

Copyright 2020 Thomas Leduc

This file is part of t4gpd.

t4gpd is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

t4gpd is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with t4gpd.  If not, see <https://www.gnu.org/licenses/>.
'''
from geopandas.geodataframe import GeoDataFrame
from numpy import arctan, cos, pi, sin
from shapely.geometry import Polygon
from t4gpd.commons.GeomLib import GeomLib
from t4gpd.commons.IllegalArgumentTypeException import IllegalArgumentTypeException
# from t4gpd.commons.RayCastingLib import RayCastingLib
from t4gpd.commons.RayCasting3Lib import RayCasting3Lib
from t4gpd.morph.geoProcesses.AbstractGeoprocess import AbstractGeoprocess


class SkyMap2D(AbstractGeoprocess):
    '''
    classdocs
    '''

    def __init__(self, buildingsGdf, nRays=180, size=4.0, maxRayLen=100.0, elevationFieldname='HAUTEUR',
                 projectionName='Stereographic'):
        '''
        Constructor
        '''
        if not isinstance(buildingsGdf, GeoDataFrame):
            raise IllegalArgumentTypeException(buildingsGdf, 'GeoDataFrame')
        self.buildingsGdf = buildingsGdf
        self.spatialIndex = buildingsGdf.sindex

        self.nRays = nRays
        # self.shootingDirs = RayCastingLib.preparePanopticRays(nRays)
        self.shootingDirs = RayCasting3Lib.preparePanopticRays(nRays)
        self.size = size
        self.maxRayLen = maxRayLen

        if elevationFieldname not in buildingsGdf:
            raise Exception('%s is not a relevant field name!' %
                            elevationFieldname)
        self.elevationFieldname = elevationFieldname

        if not projectionName in ['Stereographic']:
            raise IllegalArgumentTypeException(
                projectionName, 'spherical projection as "Stereographic"')
        self.proj = SkyMap2D.__stereographic

    @staticmethod
    def __stereographic(lat, lon, size=1.0):
        radius = (size * cos(lat)) / (1.0 + sin(lat))
        return (radius * cos(lon), radius * sin(lon))

    def runWithArgs(self, row):
        viewPoint = row.geometry.centroid

        if GeomLib.isAnIndoorPoint(viewPoint, self.buildingsGdf):
            return Polygon()

        # _, _, _, _, hws = RayCastingLib.multipleRayCast25D(self.buildingsGdf, self.spatialIndex, viewPoint,
        #                                                    self.shootingDirs, self.maxRayLen, self.elevationFieldname, True)
        _, _, _, _, hws = RayCasting3Lib.outdoorMultipleRayCast25D(self.buildingsGdf, viewPoint, self.shootingDirs, self.maxRayLen,
                                                                   self.elevationFieldname, True, h0=0.0)

        lats = [arctan(hw) for hw in hws]
        angularOffset = (2.0 * pi) / self.nRays
        lons = [i * angularOffset for i in list(range(self.nRays)) + [0]]

        pnodes = [self.proj(lats[i], lons[i], self.size)
                  for i in list(range(self.nRays)) + [0]]
        nodes = [(viewPoint.x + pnode[0], viewPoint.y + pnode[1])
                 for pnode in pnodes]

        return {'geometry': Polygon(viewPoint.buffer(self.size).exterior.coords, [nodes])}



'''
Created on 24 jul. 2023

@author: tleduc

Copyright 2020-2023 Thomas Leduc

This file is part of t4gpd.

t4gpd is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

t4gpd is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with t4gpd.  If not, see <https://www.gnu.org/licenses/>.
'''
from geopandas import GeoDataFrame
from shapely import Polygon
from t4gpd.commons.GeoDataFrameLib import GeoDataFrameLib
from t4gpd.commons.GeoProcess import GeoProcess
from t4gpd.commons.IllegalArgumentTypeException import IllegalArgumentTypeException
from t4gpd.commons.RayCasting4Lib import RayCasting4Lib


class STIsovistField2D_new(GeoProcess):
    '''
    classdocs
    '''

    def __init__(self, buildings, viewpoints, nRays=64, rayLength=100.0, withIndices=False):
        '''
        Constructor
        '''
        if not isinstance(buildings, GeoDataFrame):
            raise IllegalArgumentTypeException(
                buildings, "buildings GeoDataFrame")
        if not isinstance(viewpoints, GeoDataFrame):
            raise IllegalArgumentTypeException(
                viewpoints, "viewpoints GeoDataFrame")

        if not GeoDataFrameLib.shareTheSameCrs(buildings, viewpoints):
            raise Exception(
                "Illegal argument: buildings and viewpoints must share shames CRS!")

        self.buildings = buildings
        # CLEAN GEOMETRIES
        self.buildings.geometry = self.buildings.geometry.apply(
            lambda g: g.buffer(0))

        self.rays = RayCasting4Lib.get2DPanopticRaysGeoDataFrame(
            viewpoints, rayLength, nRays)
        self.nRays = nRays
        self.withIndices = withIndices

    def run(self):
        isovRaysField, isovField = RayCasting4Lib.multipleRayCast2D(
            self.buildings, self.rays, self.withIndices)
        return isovRaysField, isovField



'''
Created on 5 janv. 2021

@author: tleduc

Copyright 2020 Thomas Leduc

This file is part of t4gpd.

t4gpd is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

t4gpd is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with t4gpd.  If not, see <https://www.gnu.org/licenses/>.
'''
from pandas import DataFrame
from scipy.cluster.hierarchy import dendrogram, linkage
from t4gpd.commons.GeoProcess import GeoProcess
from t4gpd.commons.IllegalArgumentTypeException import IllegalArgumentTypeException

import matplotlib.pyplot as plt


class STDendrogram(GeoProcess):
    '''
    classdocs
    '''
    NUMERICS = ["int16", "int32", "int64", "float16", "float32", "float64"]

    def __init__(self, inputGdf, orientation="top", no_plot=False, verbose=True):
        '''
        Constructor
        '''
        if not isinstance(inputGdf, DataFrame):
            raise IllegalArgumentTypeException(inputGdf, "DataFrame")

        self.data = inputGdf.select_dtypes(include=self.NUMERICS)

        _fieldnames = list(self.data.columns)
        if (0 == len(_fieldnames)):
            raise Exception("There are no numeric (float) fields in the GeoDataFrame!")
        if verbose:
            print(f"The following fields are taken into account when processing: {_fieldnames}")

        if not orientation in ("bottom", "left", "right", "top"):
            raise IllegalArgumentTypeException(orientation, '("bottom", "left", "right", "top")')
        self.orientation = orientation

        if not isinstance(no_plot, bool):
            raise IllegalArgumentTypeException(no_plot, "bool")
        self.noPlot = no_plot

    def run(self):
        Z = linkage(self.data, method="ward", metric="euclidean")
        d = dendrogram(Z, orientation=self.orientation, show_contracted=True, no_plot=self.noPlot)
        plt.show()

        nClusters = len(set(d["color_list"])) - 1
        return nClusters



'''
Created on 22 aug. 2023

@author: tleduc

Copyright 2020-2023 Thomas Leduc

This file is part of t4gpd.

t4gpd is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

t4gpd is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with t4gpd.  If not, see <https://www.gnu.org/licenses/>.
'''
from geopandas import read_file
import json
from geopandas import GeoDataFrame
from rasterio.io import DatasetReader
from rasterio.mask import mask
from t4gpd.commons.IllegalArgumentTypeException import IllegalArgumentTypeException
from t4gpd.raster.AbstractRasterGeoProcess import AbstractRasterGeoProcess


class RTClip(AbstractRasterGeoProcess):
    '''
    classdocs
    '''

    def __init__(self, raster, roi, debug=False):
        '''
        Constructor
        '''
        if not isinstance(roi, GeoDataFrame):
            raise IllegalArgumentTypeException(roi, "GeoDataFrame")
        self.roi = roi

        if not isinstance(raster, DatasetReader):
            raise IllegalArgumentTypeException(raster, "DatasetReader")
        self.raster = raster
        self.debug = debug

    @staticmethod
    def __getFeatures(gdf):
        '''
        Function to parse features from GeoDataFrame in such a manner that rasterio expects them
        '''
        return [json.loads(gdf.to_json())["features"][0]["geometry"]]

    def run(self):
        out_img, out_transform = mask(
            self.raster, shapes=self.__getFeatures(self.roi), crop=True)
        out_meta = self.raster.meta.copy()
        out_meta.update({
            "driver": "GTiff",
            "height": out_img.shape[1],
            "width": out_img.shape[2],
            "transform": out_transform,
            "crs": self.roi.crs
        })
        result = self._write_and_load(out_img, out_meta, self.debug, indexes=None)
        return result


"""
roi = read_file("/home/tleduc/prj/uenv_aau_msc_course/tps/data/roi.shp")
raster = rasterio.open("/home/tleduc/data/mnt_asc/mnt_nantes_2005_1m/MNT_L93_0354_6690.asc")
r = RTClip(raster, roi).run()
print(type(r))
"""



'''
Created on 24 feb. 2021

@author: tleduc

Copyright 2020 Thomas Leduc

This file is part of t4gpd.

t4gpd is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

t4gpd is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with t4gpd.  If not, see <https://www.gnu.org/licenses/>.
'''
from t4gpd.commons.GeomLib import GeomLib
from shapely.geometry import box, MultiPolygon, Polygon


class KernelLib(object):
    '''
    classdocs
    '''

    @staticmethod
    def __getInteriorHalfPlanePolygon(a, b, bbox):
        minx, miny, maxx, maxy = bbox

        llc = [minx, miny]
        lrc = [maxx, miny]
        ulc = [minx, maxy]
        urc = [maxx, maxy]
        ap = KernelLib.__projectOntoEnvelope(a, b, bbox)
        bp = KernelLib.__projectOntoEnvelope(b, a, bbox)
        result = [bp, ap]
        
        if (a[0] == b[0]):
            if (a[1] > b[1]):
                result.append(lrc)
                result.append(urc)
            elif (a[1] < b[1]):
                result.append(ulc)
                result.append(llc)
            else:
                raise Exception('Points a and b are equals !')
        elif (a[0] < b[0]):
            if (a[1] == b[1]):
                result.append(urc)
                result.append(ulc)
            elif (a[1] < b[1]):
                if (KernelLib.__isOnTheLeftOfABSegment(a, b, urc)):
                    result.append(urc)
                result.append(ulc);
                if (KernelLib.__isOnTheLeftOfABSegment(a, b, llc)):
                    result.append(llc)
            else:  # a[1] > b[1]
                if (KernelLib.__isOnTheLeftOfABSegment(a, b, lrc)):
                    result.append(lrc)
                result.append(urc)
                if (KernelLib.__isOnTheLeftOfABSegment(a, b, ulc)):
                    result.append(ulc)
        else:  # a[0] > b[0]
            if (a[1] == b[1]):
                result.append(llc)
                result.append(lrc)
            elif (a[1] < b[1]):
                if (KernelLib.__isOnTheLeftOfABSegment(a, b, ulc)):
                    result.append(ulc)
                result.append(llc)
                if (KernelLib.__isOnTheLeftOfABSegment(a, b, lrc)):
                    result.append(lrc)
            else:  # a[1] > b[1]
                if (KernelLib.__isOnTheLeftOfABSegment(a, b, llc)):
                    result.append(llc)
                result.append(lrc)
                if (KernelLib.__isOnTheLeftOfABSegment(a, b, urc)):
                    result.append(urc)
        result.append(bp)
        return result

    @staticmethod
    def getKernel(geom):
        if ((not GeomLib.isPolygonal(geom)) or
            (GeomLib.isHoled(geom)) or
            ((isinstance(geom, MultiPolygon)) and (1 != len(geom.geoms)))):
            return None, False

        if GeomLib.isConvex(geom):
            return GeomLib.normalizeRingOrientation(geom), True

        if isinstance(geom, MultiPolygon):
            geom = geom.geoms[0]

        polygon = GeomLib.normalizeRingOrientation(geom)
        bbox = polygon.bounds

        kernel = box(*bbox)
        _coords = polygon.exterior.coords
        for i in range(1, len(_coords)):
            tmp = KernelLib.__getInteriorHalfPlanePolygon(
                _coords[i - 1], _coords[i], bbox)
            tmp = Polygon(tmp)
            kernel = kernel.intersection(tmp)
            # kernel.isEmpty() n'est pas equivalent a : 0.0 == kernel.area
            if (kernel.is_empty) or (0.0 == kernel.area):
                return None, False

        return kernel, True

    @staticmethod
    def __isOnTheLeftOfABSegment(a, b, p):
        # Test if z component of (AB x AP) >= 0
        return (0.0 <= ((b[0] - a[0]) * (p[1] - a[1]) - (b[1] - a[1]) * (p[0] - a[0])))

    @staticmethod
    def __projectOntoEnvelope(p, q, bbox):
        minx, miny, maxx, maxy = bbox
        # This method returns the projection of point p on the envelope in the
        # direction of point q.
        if (p[0] == q[0]):
            if (p[1] < q[1]):
                return [q[0], maxy]
            elif (p[1] > q[1]):
                return [q[0], miny]
            else:
                raise Exception('Points p and q are equals!');

        elif (p[0] < q[0]):
            x0 = maxx
            y0 = p[1] + ((x0 - p[0]) * (q[1] - p[1])) / (q[0] - p[0])

            if ((miny <= y0) and (y0 <= maxy)):
                return [x0, y0]
            else:
                if (p[1] == q[1]):
                    raise Exception('Should not happen !')
                    # return [x0, y0]
                elif (p[1] < q[1]):
                    y0 = maxy
                else:
                    y0 = miny
                x0 = p[0] + ((q[0] - p[0]) * (y0 - p[1])) / (q[1] - p[1])
                return [x0, y0]

        else:  # p[0] > q[0]
            x0 = minx
            y0 = p[1] + ((x0 - p[0]) * (q[1] - p[1])) / (q[0] - p[0])

            if ((miny <= y0) and (y0 <= maxy)):
                return [x0, y0]
            else:
                if (p[1] == q[1]):
                    raise Exception('Should not happen !')
                    # return [x0, y0]
                elif (p[1] < q[1]):
                    y0 = maxy
                else:
                    y0 = miny
                x0 = p[0] + ((q[0] - p[0]) * (y0 - p[1])) / (q[1] - p[1])
                return [x0, y0]



'''
Created on 24 oct. 2022

@author: tleduc

Copyright 2020-2022 Thomas Leduc

This file is part of t4gpd.

t4gpd is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

t4gpd is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with t4gpd.  If not, see <https://www.gnu.org/licenses/>.
'''
from datetime import datetime

from pandas import read_csv
from t4gpd.commons.GeoProcess import GeoProcess

from t4gpd.commons.DataFrameLib import DataFrameLib


class MeteoFranceReader(GeoProcess):
    '''
    classdocs
    '''
    FIELD_NAMES = ['station', 'timestamp', 'Tair', 'pressure_hpa', 'ws_ms', 'WindDir',
                   'ws_2_ms', 'WindDir_2', 'RH', 'insolation_min', 'global_radiation',
                   'direct_radiation', 'diffuse_radiation', 'nebulosity', 'visibility'] 

    def __init__(self, inputFile):
        '''
        Constructor
        '''
        self.inputFile = inputFile

    @staticmethod
    def interpolate(df, sensorFamily, fieldname, dt):
        _df = df[ df.sensorFamily == sensorFamily ].copy(deep=True)
        return DataFrameLib.interpolate(_df, 'timestamp', fieldname, dt)

    def run(self):
        # https://pandas.pydata.org/pandas-docs/stable/reference/api/pandas.read_csv.html
        df = read_csv(self.inputFile, decimal=',', names=self.FIELD_NAMES, sep=';',
                      skiprows=1)
        df.timestamp = df.timestamp.apply(lambda t: datetime.strptime(str(t), '%Y%m%d%H'))
        df['sensorFamily'] = df.station.apply(lambda v: f'{v}')
        return df

'''
inputFile = '/home/tleduc/prj/nm-ilots-frais/information/meteo-france-202207-FRNOR.RR22092917362047672.PPDH.KEYuUfd1B1D2x9uv92O31d1.txt'
df = MeteoFranceReader(inputFile).run()
df = df[ df.sensorFamily == 'MF-44020001'].copy(deep=True)
print(df.head(5))
# print(df)
dt = datetime(2022, 7, 10, 1, 15)
print(dt, MeteoFranceReader.interpolate(df, '44020001', 'Tair', dt))
'''



'''
Created on 28 sept. 2020

@author: tleduc

Copyright 2020 Thomas Leduc

This file is part of t4gpd.

t4gpd is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

t4gpd is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with t4gpd.  If not, see <https://www.gnu.org/licenses/>.
'''
from geopandas.geodataframe import GeoDataFrame
from numpy import mean
from t4gpd.commons.IllegalArgumentTypeException import IllegalArgumentTypeException
from t4gpd.commons.RayCasting3Lib import RayCasting3Lib
from t4gpd.morph.geoProcesses.AbstractGeoprocess import AbstractGeoprocess


class HMean(AbstractGeoprocess):
    '''
    classdocs
    '''

    def __init__(self, buildingsGdf, nRays=64, maxRayLen=100.0, elevationFieldname='HAUTEUR',
                 background=False, h0=0.0):
        '''
        Constructor
        '''
        if not isinstance(buildingsGdf, GeoDataFrame):
            raise IllegalArgumentTypeException(buildingsGdf, 'GeoDataFrame')
        self.buildingsGdf = buildingsGdf

        self.shootingDirs = RayCasting3Lib.preparePanopticRays(nRays)
        self.maxRayLen = maxRayLen

        if elevationFieldname not in buildingsGdf:
            raise Exception('%s is not a relevant field name!' % elevationFieldname)
        self.elevationFieldname = elevationFieldname
        self.background = background
        self.h0 = h0

    def runWithArgs(self, row):
        viewPoint = row.geometry.centroid

        _, _, _, hitMasks, _ = RayCasting3Lib.outdoorMultipleRayCast25D(
            self.buildingsGdf, viewPoint, self.shootingDirs,
            self.maxRayLen, self.elevationFieldname, self.background, self.h0)

        hitHeights = [0.0 if f is None else f[self.elevationFieldname] for f in hitMasks]

        return {
            'hmean': float(mean(hitHeights))
            }



'''
Created on 10 sept. 2021

@author: tleduc

Copyright 2020-2021 Thomas Leduc

This file is part of t4gpd.

t4gpd is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

t4gpd is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with t4gpd.  If not, see <https://www.gnu.org/licenses/>.
'''
from t4gpd.commons.GeoProcess import GeoProcess

from qgis.PyQt.QtCore import QRectF
from qgis.PyQt.QtGui import QFont
from qgis.core import (QgsLayoutExporter, QgsLayoutItemLabel, QgsLayoutItemLegend,
                       QgsLayoutItemMap, QgsLayoutItemScaleBar,
                       QgsLayoutPoint, QgsLayoutSize, QgsLegendStyle,
                       QgsPrintLayout, QgsProject,
                       QgsUnitTypes)


class PdfMapWriter(GeoProcess):
    '''
    classdocs
    '''

    def __init__(self, outputFileName, layerExtent, xOffset=0, yOffset=0,
                 legendXOffset=238, legendYOffset=1, scaleBarXOffset=60,
                 scaleBarYOffset=120, scaleBarNSegm=3, scalebarBg=False,
                 fileFormat='a4landscape', textLabel=None, labelXOffset=10,
                 labelYOffset=10, fontLabel=QFont('Cambria', 34, QFont.Bold),
                 numUnitsPerSegment=None, scaleBarFontSize=QFont('Cambria', 24),
                 fontLegend=None, symbolHeight=None, legendBg=None):
        '''
        Constructor
        '''
        self.outputFileName = outputFileName
        _layout = PdfMapWriter.__getLayout(layerExtent, xOffset, yOffset,
            legendXOffset, legendYOffset,
            scaleBarXOffset, scaleBarYOffset, scaleBarNSegm, scalebarBg,
            fileFormat, textLabel, labelXOffset, labelYOffset,
            fontLabel, numUnitsPerSegment, scaleBarFontSize,
            fontLegend, symbolHeight, legendBg)
        self.layoutExporter = QgsLayoutExporter(_layout)

    @staticmethod
    def __getLayout(layerExtent, xOffset=0, yOffset=0, legendXOffset=238, legendYOffset=1, scaleBarXOffset=60,
        scaleBarYOffset=120, scaleBarNSegm=3, scalebarBg=False, fileFormat='a4landscape',
        textLabel=None, labelXOffset=10, labelYOffset=10, fontLabel=QFont('Cambria', 34, QFont.Bold),
        numUnitsPerSegment=None, scaleBarFontSize=QFont('Cambria', 24), fontLegend=None, symbolHeight=None,
        legendBg=None):
        if 'a4landscape' == fileFormat:
            width, height = 297, 210
        elif 'a4portrait' == fileFormat:
            width, height = 210, 297
        elif 'square' == fileFormat:
            width, height = 210, 210
        layoutSize = QgsLayoutSize(width=width, height=height, units=QgsUnitTypes.LayoutMillimeters)

        project = QgsProject.instance()
        layout = QgsPrintLayout(project)
        layout.initializeDefaults()
        layout.pageCollection().page(0).setPageSize(layoutSize)

        # CANVAS
        myMap = QgsLayoutItemMap(layout)
        myMap.setRect(QRectF(layerExtent.xMinimum(), layerExtent.yMinimum(), layerExtent.xMaximum(), layerExtent.yMaximum()))
        myMap.setExtent(layerExtent)
        layout.addItem(myMap)
        myMap.attemptResize(layoutSize)

        # TITLE
        if textLabel is not None:
            label = QgsLayoutItemLabel(layout)
            label.setText(textLabel)
            label.setFont(fontLabel)
            label.adjustSizeToText()
            layout.addItem(label)
            label.attemptMove(QgsLayoutPoint(labelXOffset, labelYOffset, QgsUnitTypes.LayoutMillimeters))
            label.setBackgroundEnabled(1)

        # LEGEND
        legend = QgsLayoutItemLegend(layout)
        legend.setTitle('Legend')
        legend.setLinkedMap(myMap)
        if fontLegend is not None:
            legend.setStyleFont(QgsLegendStyle.Title, fontLegend)
            legend.setStyleFont(QgsLegendStyle.Subgroup, fontLegend)
            legend.setStyleFont(QgsLegendStyle.SymbolLabel, fontLegend)
        if symbolHeight is not None:
            legend.setSymbolHeight(symbolHeight)
        layout.addItem(legend)
        legend.attemptMove(QgsLayoutPoint(legendXOffset, legendYOffset, QgsUnitTypes.LayoutMillimeters))
        # ## legend.setFrameEnabled(True)
        if (legendBg is not None) and legendBg:
            legend.setBackgroundEnabled(1)
        # legend.setBackgroundEnabled(1 if scalebarBg else 0)

        # SCALE BAR
        scaleBar = QgsLayoutItemScaleBar(layout)
        # Style possibilities are: 'Single Box', 'Double Box', 'Line Ticks Middle', 'Line Ticks Down', 'Line Ticks Up', 'Numeric'
        scaleBar.setStyle('Single Box')
        scaleBar.setLinkedMap(myMap)
        scaleBar.applyDefaultSize()
        scaleBar.setNumberOfSegmentsLeft(0)
        scaleBar.setNumberOfSegments(scaleBarNSegm)
        if numUnitsPerSegment is not None:
            scaleBar.setUnitsPerSegment(numUnitsPerSegment)
        scaleBar.setFont(scaleBarFontSize)
        layout.addItem(scaleBar)
        scaleBar.attemptMove(QgsLayoutPoint(scaleBarXOffset, scaleBarYOffset, QgsUnitTypes.LayoutMillimeters))
        # ## scaleBar.setFrameEnabled(True)
        scaleBar.setBackgroundEnabled(1 if scalebarBg else 0)

        return layout

    def run(self):
        self.layoutExporter.exportToPdf(self.outputFileName,
                                        QgsLayoutExporter.PdfExportSettings())



'''
Created on 27 sept. 2020

@author: tleduc

Copyright 2020 Thomas Leduc

This file is part of t4gpd.

t4gpd is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

t4gpd is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with t4gpd.  If not, see <https://www.gnu.org/licenses/>.
'''
from geopandas.geodataframe import GeoDataFrame
from shapely.geometry import box
from shapely.ops import  linemerge, unary_union
from shapely.prepared import prep
from t4gpd.commons.GeoProcess import GeoProcess
from t4gpd.commons.GeomLib import GeomLib
from t4gpd.commons.IllegalArgumentTypeException import IllegalArgumentTypeException
from t4gpd.morph.STPointsDensifier import STPointsDensifier
from t4gpd.morph.STVoronoiPartition import STVoronoiPartition


class STSkeletonizeTheVoid(GeoProcess):
    '''
    classdocs
    '''

    def __init__(self, buildingsGdf, samplingDist=10.0):
        '''
        Constructor
        '''
        if not isinstance(buildingsGdf, GeoDataFrame):
            raise IllegalArgumentTypeException(buildingsGdf, 'GeoDataFrame')
        self.buildingsGdf = buildingsGdf
        self.spatialIdx = buildingsGdf.sindex

        self.samplingDist = samplingDist

    def run(self):
        roiGeom = box(*self.buildingsGdf.total_bounds)

        # Sample the building contours
        _nodesGdf = STPointsDensifier(self.buildingsGdf, self.samplingDist).run()

        # Voronoi Diagram
        _voronoiGdf = STVoronoiPartition(_nodesGdf).run()

        pgeom = prep(unary_union(self.buildingsGdf.geometry))
        lls = []
        for geom in _voronoiGdf.geometry:
            lls.extend(filter(pgeom.disjoint, GeomLib.toListOfBipointsAsLineStrings(geom)))
        lls = linemerge(lls).intersection(roiGeom)

        return GeoDataFrame([{'gid': 1, 'geometry': lls}], crs=self.buildingsGdf.crs)



'''
Created on 25 nov. 2022

@author: tleduc

Copyright 2020-2022 Thomas Leduc

This file is part of t4gpd.

t4gpd is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

t4gpd is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with t4gpd.  If not, see <https://www.gnu.org/licenses/>.
'''
from geopandas import GeoDataFrame
from t4gpd.commons.GeoProcess import GeoProcess
from t4gpd.commons.IllegalArgumentTypeException import IllegalArgumentTypeException
from t4gpd.commons.sun.PySolarSunLib import PySolarSunLib


class ExtraProcessing(GeoProcess):
    '''
    classdocs
    '''

    def __init__(self, dfImuMob, inplace=False):
        '''
        Constructor
        '''
        if not isinstance(dfImuMob, GeoDataFrame):
            raise IllegalArgumentTypeException(dfImuMob, "GeoDataFrame")
        if inplace:
            self.dfImuMob = dfImuMob
        else:
            self.dfImuMob = dfImuMob.copy(deep=True)
        self.crs = dfImuMob.crs

        self.sunLib = PySolarSunLib(dfImuMob)

        assert dfImuMob.timestamp.is_monotonic_increasing, f"dfImu.timestamp is not monotonic increasing!"

    def run(self):
        self.dfImuMob.to_crs("epsg:4326", inplace=True)
        self.dfImuMob["lat_cor"] = self.dfImuMob.geometry.apply(lambda geom: geom.y)
        self.dfImuMob["lon_cor"] = self.dfImuMob.geometry.apply(lambda geom: geom.x)
        self.dfImuMob.to_crs(self.crs, inplace=True)

        self.dfImuMob["epoch"] = self.dfImuMob.timestamp.apply(
            lambda dt: dt.timestamp())
            # lambda dt: int(dt.strftime("%s")))
        self.dfImuMob.epoch = self.dfImuMob.epoch.astype(int)

        self.dfImuMob["sun_alti"], self.dfImuMob["sun_azim"] = self.dfImuMob.apply(
            lambda row: self.sunLib.getSolarAnglesInDegrees(row.timestamp.to_pydatetime()),
            axis=1, result_type="expand").to_numpy().transpose()

        return self.dfImuMob
